<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 642086b77e15 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-04-16 02:04 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * Customer filter (model) for document (or object) filters
 *
 */
class we_customer_documentFilter extends we_customer_abstractFilter{
	const ACCESS = 'f_1';
	const CONTROLONTEMPLATE = 'f_2';
	const NO_ACCESS = 'f_3';
	const NO_LOGIN = 'f_4';

	/**
	 * Flag if access control is made by template or not
	 *
	 * @var boolean
	 */
	private $accessControlOnTemplate = false;

	/** @var int Id of Document which is shown when customer is not logged in 	 */
	private $errorDocNoLogin = 0;

	/** @var int Id of Document which is shown when customer has no acces	 */
	private $errorDocNoAccess = 0;

	/**
	 * @param string $modelTable
	 * @param integer $modelId
	 * @param boolean $accessControlOnTemplate
	 * @param integer $errorDocNoLogin
	 * @param integer $errorDocNoAccess
	 * @param integer $mode
	 * @param int[] $specificCustomers
	 * @param array<int,array{operation:int,field:string,value:string,logic:string}> $filter
	 * @param int[] $whiteList
	 * @param int[] $blackList
	 */
	public function __construct(string $modelTable, int $modelId = 0, bool $accessControlOnTemplate = true, int $errorDocNoLogin = 0, int $errorDocNoAccess = 0, int $mode = self::OFF, array $specificCustomers = [
		], array $filter = [], array $whiteList = [], array $blackList = []){
		parent::__construct($modelTable, $modelId, $mode, $specificCustomers, $blackList, $whiteList, $filter);

		$this->accessControlOnTemplate = $accessControlOnTemplate;
		$this->errorDocNoLogin = $errorDocNoLogin;
		$this->errorDocNoAccess = $errorDocNoAccess;
	}

	/**
	 * initializes and returns filter object from db object. Called after $db->query();
	 *
	 * @return we_customer_documentFilter
	 */
	public static function getFilterByDbHash(array $hash): self{
		return new self($hash['modelTable'], intval($hash['modelId']), boolval($hash['accessControlOnTemplate']), intval($hash['errorDocNoLogin']), intval($hash['errorDocNoAccess']), intval($hash['mode']), we_base_request::filterIntList($hash['specificCustomers']), we_unserialize($hash['filter']), we_base_request::filterIntList($hash['whiteList']), we_base_request::filterIntList($hash['blackList']));
	}

	/**
	 * initializes and returns filter object from request
	 *
	 * param webeditionDocument or objectFile
	 * @return we_customer_documentFilter
	 */
	public function getCustomerFilterFromRequest(int $id): void{
		parent::getCustomerFilterFromRequest($id);
		$mode = $this->getMode();
		$this->accessControlOnTemplate = ($mode == self::OFF ? true : (we_base_request::string('wecf_accessControlOnTemplate') === 'onTemplate'));
		$this->errorDocNoLogin = ($mode == self::OFF ? 0 : we_base_request::int('wecf_noLoginId', 0));
		$this->errorDocNoAccess = ($mode == self::OFF ? 0 : we_base_request::int('wecf_noAccessId', 0));
	}

	/**
	 * get additional condition for listviews
	 *
	 * @return string
	 */
	static function getConditionForListviewQuery(string $filter, we_listview_base $obj, int $classID = 0, array $ids = []): string{
		if($filter === 'off' || $filter === 'false' || $filter === 'all' || empty($filter)){
			return '';
		}
		if(!self::customerIsLogedIn()){ //we don't show any documents with an customerfilter
			if($obj instanceof we_listview_search){ // search
				//FIXME: changed tblINDEX!
				return ' AND (
(i.ClassID=0 AND i.ID NOT IN(SELECT modelId FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelTable="' . stripTblPrefix(FILE_TABLE) . '"))
	OR
(i.ClassID>0 AND i.ID NOT IN(SELECT modelId FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelTable="' . stripTblPrefix(OBJECT_FILES_TABLE) . '")) )';
			}
			return ($classID ?
				' AND of.ID NOT IN(SELECT modelId FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelTable="' . stripTblPrefix(OBJECT_FILES_TABLE) . '")' :
				' AND f.ID NOT IN(SELECT modelId FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelTable="' . stripTblPrefix(FILE_TABLE) . '")'
				);
		}

		// if customer is not logged in, all documents/objects with filters must be hidden
		$restrictedFilesForCustomer = self::getFilesWithRestrictionsOfCustomer($obj, $filter, $classID, $ids);

		if($obj instanceof we_listview_search){ // search
			$queryTail = [];
			// build query from restricted files, regard search and normal listview
			foreach($restrictedFilesForCustomer as $tab => $fileArray){
				if($fileArray){
					$queryTail [] = '(' . ($tab === 'tblObjectFiles' ? ' ClassID>0' : 'ClassID=0') . ' AND ID NOT IN(' . implode(', ', $fileArray) . '))';
				}
			}
			return ' AND ' . implode(' OR ', $queryTail);
		}
		$queryTail = '';
		// build query from restricted files, regard search and normal listview
		foreach($restrictedFilesForCustomer as $tab => $fileArray){
			if($fileArray){
				$queryTail .= ' AND ' .
					($classID && $tab === 'tblObjectFiles' ? 'of.ID' : 'f.ID') .
					' NOT IN(' . implode(', ', $fileArray) . ')';
			}
		}

		return $queryTail;
	}

	/**
	 * compares two filters and returns true if they have equal data
	 * @param string $table
	 * @param int $id1
	 * @param int $id2
	 * @return bool
	 */
	public static function filterAreQual(string $table, int $id1, int $id2): bool{
		$all = DB_WE::$inst->getAllq('SELECT * FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelTable="' . DB_WE::$inst->escape(stripTblPrefix($table)) . '" AND modelId IN (' . $id1 . ',' . $id2 . ')');
		'@phan-var array<int,array<string,string|int>> $all';
		$filter1 = $all[0] ?? [];
		$filter2 = $all[1] ?? [];
		unset($filter1['modelId'], $filter2['modelId']);
		return empty(array_diff_assoc($filter1, $filter2));
	}

	/**
	 * gets the right error document id
	 *
	 * @param String $errorConstant
	 * @return integer
	 */
	function getErrorDoc(string $errorConstant): int{
		switch($errorConstant){
			case self::NO_LOGIN:
				return ($this->errorDocNoLogin ?: $this->errorDocNoAccess);
			case self::NO_ACCESS:
				return ($this->errorDocNoAccess ?: $this->errorDocNoLogin);
			default:
				return 0;
		}
	}

	/**
	 * saves the filter data in db. Call this on save method of model
	 *
	 * param webeditionDocument or objectFile
	 * @param array<string,mixed> $set
	 */
	public function save(array $set = []): void{
		if(!$this->getModelId()){
			return;
		}
		$pre = getHash('SELECT * FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelTable="' . DB_WE::$inst->escape($this->getModelTable()) . '" AND modelId=' . $this->getModelId());
		parent::save([
			'accessControlOnTemplate' => $this->getAccessControlOnTemplate(),
			'errorDocNoLogin' => $this->getErrorDocNoLogin(),
			'errorDocNoAccess' => $this->getErrorDocNoAccess(),
		]);

		$post = getHash('SELECT * FROM ' . CUSTOMER_FILTER_TABLE . ' WHERE modelTable="' . DB_WE::$inst->escape($this->getModelTable()) . '" AND modelId=' . $this->getModelId());
		//if the filter changed, find all nav entries with this document, and invaldate the cache for them
		if(array_diff_assoc($pre, $post)){
			$ids = DB_WE::$inst->getAllq('SELECT ID FROM ' . NAVIGATION_TABLE . ' WHERE 	UseDocumentFilter=1 AND
	SelectionType="' . (addTblPrefix($this->getModelTable()) === OBJECT_FILES_TABLE ? we_navigation_navigation::STYPE_OBJLINK : we_navigation_navigation::STYPE_DOCLINK) . '" AND
	LinkID=' . $this->getModelId(), true);
			'@phan-var int[] $ids';

			foreach($ids as $id){
				we_navigation_cache::delNavigationTree($id);
			}
		}
	}

	/**
	 * private function. gets all file ids which customer can not accesss
	 *
	 * @param we_listview_base $obj
	 * @return array
	 */
	private static function getFilesWithRestrictionsOfCustomer(we_listview_base $obj, string $queryfilter, int $classID, array $ids): array{
		//FIXME: this will query ALL documents with restrictions - this is definately not what we want!
		$cid = !empty($_SESSION['webuser']['registered']) && $_SESSION['webuser']['ID'] ? $_SESSION['webuser']['ID'] : 0;
		//cache result
		static $filesWithRestrictionsForCustomer = [];

		// detect all files/objects with restrictions
		list($join, $cond) = $obj->getCustomerRestrictionQuery($classID);

// if customer is not logged in=> return NO_LOGIN
		// else return correct filter
		// execute the query (get all existing filters)
		$query = 'SELECT cf.*
FROM ' .
			CUSTOMER_FILTER_TABLE . ' cf ' . $join . '
WHERE ' . $cond . '
cf.mode IN(' . self::FILTER . ',' . self::SPECIFIC . ') AND
(
	(cf.mode=' . self::FILTER . ' AND !FIND_IN_SET(' . $cid . ',cf.whiteList) ) OR
	(cf.mode=' . self::SPECIFIC . ' AND !FIND_IN_SET(' . $cid . ',cf.specificCustomers))
)' . ($ids ?
			' AND cf.modelId IN (' . implode(',', $ids) . ')' :
			''
			);
		$key = md5($query);
		if(isset($filesWithRestrictionsForCustomer[$key])){
			return $filesWithRestrictionsForCustomer[$key];
		}

		$db = new DB_WE();
// visitor is logged in
		$filesWithRestrictionsForCustomer[$key] = [];
		foreach($db->getAllq($query) as $record){
			'@phan-var array $record';
			$filter = self::getFilterByDbHash($record);
			$perm = $filter->accessForVisitor($filter->getModelId(), false, true);
			switch($perm){
				case self::NO_ACCESS:
				case self::NO_LOGIN:
					$filesWithRestrictionsForCustomer[$key][$filter->getModelTable()][] = $filter->getModelId();
					break;
				case self::CONTROLONTEMPLATE:
					if($queryfilter === 'all' || $queryfilter === 'true'){
						$filesWithRestrictionsForCustomer[$key][$filter->getModelTable()][] = $filter->getModelId();
					}
					break;
				case self::ACCESS:
					break;
			}
		}
		return $filesWithRestrictionsForCustomer[$key];
	}

	/**
	 * checks if visitor has acces to see the document or object
	 *
	 * @return string
	 */
	function accessForVisitor(int $id, bool $fromIfRegisteredUser = false, bool $fromListviewCheck = false): string{
		if($id !== $this->getModelId()){//not for this model, allow access
			return self::ACCESS;
		}
		// model is correct
		if(!$fromListviewCheck && $this->getAccessControlOnTemplate() && !$fromIfRegisteredUser){
			// access control is on template (for we:ifregisteredUser)
			return self::CONTROLONTEMPLATE;
		}

		if($this->getMode() !== self::OFF && !self::customerIsLogedIn()){ // no customer logged in
			return self::NO_LOGIN;
		}

		if(!$this->customerHasAccess()){
			return self::NO_ACCESS;
		}

		return self::ACCESS;
	}

	/**
	 * Accessor method for $this->_accessControlOnTemplate
	 *
	 * @return boolean
	 */
	function getAccessControlOnTemplate(): bool{
		return $this->accessControlOnTemplate;
	}

	/**
	 * Accessor method for $this->_errorDocNoLogin
	 *
	 * @return integer
	 */
	function getErrorDocNoLogin(): int{
		return $this->errorDocNoLogin;
	}

	/**
	 * Accessor method for $this->_errorDocNoAccess
	 *
	 * @return integer
	 */
	function getErrorDocNoAccess(): int{
		return $this->errorDocNoAccess;
	}

	/**
	 * initializes and returns filter object
	 *
	 * @param integer $id
	 * @param string $table
	 * @return we_customer_documentFilter
	 */
	public static function getFilterByIdAndTable(int $id, string $table): we_customer_documentFilter{
		$hash = parent::getFilterHashFromDb($id, NAVIGATION_TABLE);
		return $hash ? self::getFilterByDbHash($hash) : new self($table, $id);
	}

}
