<?php
declare(strict_types=1);
/**
 * webEdition CMS
 *
 * $Rev: aacab26fe68a $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-05-05 10:25 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * simplified representation of the navigation item
 */
class we_navigation_item{
	/** @var int */
	var $id;

	/** @var string */
	var $icon;

	/** @var int */
	protected $docid;

	/** @var string */
	var $table;

	/** @var int */
	protected $parentid;

	/** @var string */
	var $text;

	/** @var string */
	var $display;

	/** @var string */
	var $name;

	/** @var string */
	var $href;

	/** @var string */
	var $type;

	/** @var int */
	var $level;

	/** @var int */
	var $position;

	/** @var array<int,int> */
	static $currentPosition = [];

	/** @var bool */
	var $current = false;

	/** @var bool */
	var $containsCurrent = false;

	/** @var bool */
	var $CurrentOnUrlPar;

	/** @var bool */
	var $CurrentOnAnker;

	/** @var bool */
	var $currentOnCat;

	/** @var string */
	var $catParam;

	/** @var array<string,string> */
	var $attributes;

	/** @var string */
	var $title;

	/** @var string */
	var $anchor;

	/** @var string */
	var $target;

	/** @var string */
	var $lang;

	/** @var string */
	var $hreflang;

	/** @var string */
	var $accesskey;

	/** @var string */
	var $tabindex;

	/** @var string */
	var $rel;

	/** @var we_navigation_customerFilter|null */
	var $customers = null;

	/** @var array */
	var $items = [];

	/** @var bool|null */
	private $visible = null;

	/** @var bool */
	private $linkValid = true;

	/** @var bool */
	private $customerAccess = true;

	/**
	 * we_navigation_item constructor.
	 * @param $id
	 * @param $docid
	 * @param $table
	 * @param $text
	 * @param $display
	 * @param $href
	 * @param string $type
	 * @param string $icon
	 * @param array<string,string> $attributes
	 * @param ?we_navigation_customerFilter $customers
	 * @param bool $CurrentOnUrlPar
	 * @param bool $CurrentOnAnker
	 * @param bool $currentOnCat
	 * @param string $catParam
	 */
	public function __construct(int $id, int $docid, string $table, string $text, string $display, string $href, string $type, string $icon, array $attributes, we_navigation_customerFilter $customers = null, bool $CurrentOnUrlPar = false, bool $CurrentOnAnker = false, bool $currentOnCat = false, string $catParam = ''){
		$this->id = $id;
		$this->parentid = 0;
		$this->name = $text;
		$this->text = (empty($display) ? $text : $display);
		$this->display = $display;
		$this->docid = $docid;
		$this->table = $table;
		$this->href = $href;
		$this->type = $type;
		$this->icon = $icon;
		$this->level = 0;
		$this->position = 0;
		$this->CurrentOnUrlPar = $CurrentOnUrlPar;
		$this->CurrentOnAnker = $CurrentOnAnker;
		$this->currentOnCat = $currentOnCat;
		$this->catParam = $catParam;

		$this->attributes = $attributes;
		$this->title = $attributes['title'] ?? '';
		$this->anchor = $attributes['anchor'] ?? '';
		$this->target = $attributes['target'] ?? '';
		$this->lang = $attributes['lang'] ?? '';
		$this->hreflang = $attributes['hreflang'] ?? '';
		$this->accesskey = $attributes['accesskey'] ?? '';
		$this->tabindex = $attributes['tabindex'] ?? '';
		$this->rel = $attributes['rel'] ?? '';

		$this->customers = $customers;
		switch($this->table){
			case FILE_TABLE:
				//in case the docid is 0, we assume this a structural element, which has an "valid link"
				$this->linkValid = (!$this->docid) || ( fB('SELECT 1 FROM ' . FILE_TABLE . ' f WHERE f.ID=' . $this->docid . ' AND f.IsPublished=1'));
				break;
			// #6916
			case OBJECT_FILES_TABLE :
				$this->linkValid = fB('SELECT 1 FROM ' . OBJECT_FILES_TABLE . ' `of` WHERE of.ID=' . $this->docid . ' AND of.IsPublished=1');
				break;
			default:
				//this is external url, or anything we can't test for
				$this->linkValid = true;
		}
	}

	public function __wakeup(): void{
		//need to reset customer access if the object was serialized (in cache)
		$this->customerAccess = true;
		$this->visible = null;
		//leave linkValid, since on publish the cache is regenerated
	}

	/**
	 * @param $item
	 */
	function addChild(we_navigation_item $item): void{
		$item->parentid = $this->id;
		$item->level = $this->level + 1;
		$this->items['id' . $item->id] = &$item;
		$item->position = count($this->items);
	}

	/**
	 * @param we_navigation_items $weNavigationItems
	 */
	function setCurrent(we_navigation_items $weNavigationItems): void{
		$this->current = true;

		if($this->level != 0 && ($item = $weNavigationItems->getItem('id' . $this->parentid))){
			$item->setCurrent($weNavigationItems);
			$this->containsCurrent = true;
		}
	}

	/**
	 * @param we_navigation_items $weNavigationItems
	 * @param bool $self
	 */
	function unsetCurrent(we_navigation_items $weNavigationItems, bool $self = true): void{
		if($self){
			$this->current = false;
		}

		if($this->level != 0 && $weNavigationItems->getItem('id' . $this->parentid)){
			foreach($this->items as $i){
				$i->unsetCurrent($weNavigationItems);
			}
			$this->containsCurrent = false;
		}
	}

	public function getParentId(): int{
		return $this->parentid;
	}

	private function getCurrentRequestID(): int{
//FIXME do we need this any more since we_base_WE::$mainDoc['doc'] == we_base_WE::$we_obj in case of OBJECT_FILES_TABLE ??
		static $cleanRequestUri = false;
		switch($this->table){
			case OBJECT_FILES_TABLE:
				try{
					$obj = we_base_WE::getWeObj();
				} catch (we_exception_base $e){
					return 0;
				}

				if($cleanRequestUri === false){
					$urlLookingFor = (!empty($_SERVER['REDIRECT_URL']) && !strpos($_SERVER['REDIRECT_URL'], ltrim(WEBEDITION_DIR, '/'))) ?
						urldecode($_SERVER['REDIRECT_URL']) :
						(isset($_SERVER['REQUEST_URI']) && !strpos($_SERVER['REQUEST_URI'], ltrim(WEBEDITION_DIR, '/')) ?
						(string) (parse_url(urldecode($_SERVER['REQUEST_URI']), PHP_URL_PATH) ?: '' ) :
						'');
					$path_parts = pathinfo($urlLookingFor);
					$cleanRequestUri = defined('WE_REDIRECTED_SEO') ? WE_REDIRECTED_SEO : //Fix #11057
						(isset($_SERVER['REQUEST_URI']) ? //Fix #11246
						rtrim((NAVIGATION_DIRECTORYINDEX_HIDE && seoIndexHide($path_parts['basename']) ? $path_parts['dirname'] : $urlLookingFor), '/') :
						'');
				}
				return (isset($_SERVER['REQUEST_URI']) && (empty($cleanRequestUri) || stripos($this->href, $cleanRequestUri) !== false) ? $obj->ID : 0);
			case FILE_TABLE:
				return (we_base_WE::$mainDoc['doc'] && (!(we_base_WE::$mainDoc['doc'] instanceof we_contents_objectFile)) ? we_base_WE::$mainDoc['doc']->ID : 0);
			default:
				return 0;
		}
	}

	/**
	 * @param we_navigation_items $weNavigationItems
	 * @return bool
	 */
	function isCurrent(we_navigation_items $weNavigationItems): bool{
		$id = $this->getCurrentRequestID();
		if($id && ($this->docid == $id)){
			static $uri = null;
			static $uriarrq = [];
			$refarrq = [];

			$uri = ($uri === null ? parse_url(str_replace('&amp;', '&', $_SERVER['REQUEST_URI'])) : $uri);
			$ref = parse_url(str_replace('&amp;', '&', $this->href));
			if(!empty($uri['query']) && !$uriarrq){
				parse_str($uri['query'], $uriarrq);
			}
			if(!empty($ref['query'])){
				parse_str($ref['query'], $refarrq);
			}
			if(($this->CurrentOnAnker || $this->currentOnCat) && !$this->CurrentOnUrlPar){
				//remove other param tha "anchors" or catParams respectively
				$tmpUriarrq = $tmpRefarrq = [];
				if($this->CurrentOnAnker){
					$tmpUriarrq['we_anchor'] = $uriarrq['we_anchor'] ?? '#';
					$tmpRefarrq['we_anchor'] = $refarrq['we_anchor'] ?? '#';
				}
				if($this->currentOnCat){
					$tmpUriarrq[$this->catParam] = $uriarrq[$this->catParam] ?? '#';
					$tmpRefarrq[$this->catParam] = $refarrq[$this->catParam] ?? '#';
				}
			} else {
				$tmpUriarrq = $uriarrq;
				$tmpRefarrq = $refarrq;
			}
			$allfound = true;
			//current is true, if all arguements set in navigation match current request - if we have more (maybe a form, etc.) ignore this.
			foreach($tmpRefarrq as $key => $val){
				$allfound = $allfound && isset($tmpUriarrq[$key]) && $tmpUriarrq[$key] == $val;
			}

			if($allfound){
				$this->setCurrent($weNavigationItems);
			} elseif($this->current){
				$this->unsetCurrent($weNavigationItems);
			}
			return $allfound;
		}

		if($id && ($this->docid == $id) && !($this->CurrentOnUrlPar || $this->CurrentOnAnker)){
			$this->setCurrent($weNavigationItems);
			return true;
		}

		if($this->current){
			$this->unsetCurrent($weNavigationItems);
		}
		return false;
	}

	/**
	 * @return bool|int
	 */
	public function isVisible(): bool{
		if($this->visible != null){
			//item is determined
			return (bool) $this->visible;
		}
		$visible = $this->linkValid;

		if($this->customers && $this->customers->getMode() != we_customer_abstractFilter::OFF){ // only init filter if access is limited
			$this->customerAccess = $this->customers->customerHasAccess();
			$visible = $visible && $this->customerAccess;
		}
		return ($this->visible = $visible);
	}

	public function getLastVisibleItemID(): int{
		static $lasts = [];
		if(isset($lasts[$this->id])){
			return $lasts[$this->id];
		}
		for(end($this->items); current($this->items); prev($this->items)){
			if(current($this->items)->isVisible()){
				return ($lasts[$this->id] = current($this->items)->id);
			}
		}
		return ($lasts[$this->id] = -1);
	}

	public function setLevel(): void{
		self::$currentPosition[$this->level] = 0;
	}

	/**
	 * @param $weNavigationItems
	 * @param int $depth
	 */
	function writeItem(we_navigation_items $weNavigationItems, int $depth = 0): void{
		if(!isset(self::$currentPosition[$this->level])){
			self::$currentPosition[$this->level] = 0;
		}
		if(!($depth === 0 || $this->level <= $depth)){
			return;
		}
		if(!$this->isVisible()){
			if($this->type === we_navigation_items::TYPE_ITEM || !$this->customerAccess){
				//in case of an item, if this is not visible, we are finished
				//or if the folder/root is protected by customer access
				return;
			}
			//in case of folder: check if there are visible subelements
			$vsub = false;
			//we check only one level
			foreach($this->items as $item){
				if($item->isVisible()){
					$vsub = true;
					break;
				}
			}
			if(!$vsub){
				return;
			}
		}
		we_tag_navigationWrite::$navObj->weNavigationItemArray[] = &$this;
		//use this since items might be invisible
		self::$currentPosition[$this->level]++;
		$class = we_tag_navigationWrite::$navObj->funcPrefix;
		$func = $weNavigationItems->getTemplate($this);

		$class::$func();

		array_pop(we_tag_navigationWrite::$navObj->weNavigationItemArray);
	}

	/**
	 * @param $attribs
	 * @return string
	 */
	public function getNavigationField(array $attribs): string{
		$fieldname = we_weTag_tag::getAttribute('_name_orig', $attribs, '', we_base_request::STRING);
		$compl = we_weTag_tag::getAttribute('complete', $attribs, '', we_base_request::STRING);
		$fields = we_weTag_tag::getAttribute('attributes', $attribs, [], we_base_request::STRING_LIST);
		unset($attribs['complete'], $attribs['_name_orig'], $attribs['attributes']);
		// name
		if($fieldname){
			//$fieldname = $fieldname === 'aria-label' ? 'title' : $fieldname;
			$val = (!empty($this->{$fieldname}) ?
				$this->{$fieldname} :
				(!empty($this->attributes[$fieldname]) ?
				$this->attributes[$fieldname] :
				''));
			switch($fieldname){
				case 'title':
					return htmlspecialchars($val);
				case 'href':
					return $this->linkValid ? $val : '';
				default:
					return (string) $val;
			}
		}
		$attribs = $this->getNavigationFieldAttributes($attribs, $compl ? [$compl] : $fields);
		/* if(!empty($attribs['title'])){
		  $attribs['aria-label'] = $attribs['title'];
		  } */

		// complete
		switch($compl){
			case 'link':
				if(empty($this->text)){
					return '';
				}
				return (empty($attribs['href']) || !$this->linkValid ? $this->text : we_weTag_tag::htmlTag('a', $attribs, $this->text));
			case 'image':
				if(empty($this->icon) || $this->icon == '/'){
					return '';
				}
				return we_weTag_tag::htmlTag('img', $attribs);
			default:
				break;
		}

		// attributes
		$code = '';
		foreach($fields as $key){
			switch($key){
				case 'href':
					if(!$this->linkValid){
						break;
					}
				default:
					$value = empty($attribs[$key]) ? '' : $attribs[$key];
					$code .= ' ' . ($key === 'link_attribute' ? $value : $key . '="' . $value . '"');
			}
		}
		return $code;
	}

	/**
	 * @param $attribs
	 * @return array
	 */
	private function getNavigationFieldAttributes(array $attribs, array $fields): array{
		foreach($fields as $field){
			switch($field){
				case 'link' :
					$useFields = [
						'href',
						'title',
						'target',
						'lang',
						'hreflang',
						'accesskey',
						'tabindex',
						'rel',
						'link_attribute'
					];
					foreach($useFields as $curField){
						if(!empty($this->{$curField})){
							$attribs[$curField] = ($curField === 'title' ?
								htmlspecialchars($this->{$curField}) :
								$this->{$curField});
						} elseif(!empty($this->attributes[$curField])){
							$attribs[$curField] = ($curField === 'link_attribute' ? // Bug #3741
								$this->attributes[$curField] :
								htmlspecialchars($this->attributes[$curField]));
						}
					}
					break;
				case 'image' :
					$iconid = path_to_id($this->icon, FILE_TABLE);
					if($iconid){
						$attribs['src'] = $this->icon;
						$useFields = ['width', 'height', 'border', 'hspace', 'vspace', 'align', 'alt', 'title'];
						foreach($useFields as $curField){
							if(!empty($this->attributes['icon_' . $curField])){
								$attribs[$curField] = $this->attributes['icon_' . $curField];
							}
						}
						$imgObj = new we_document_image($iconid);
						if(($rollOverClass = $imgObj->getRollOverClass())){
							$attribs['class'] = (empty($attribs['class']) ? '' : $attribs['class'] . ' ') . $rollOverClass;
						}
					}
					break;
				default:
					if(!empty($this->{$field})){
						$attribs[$field] = htmlspecialchars($this->{$field});
					} elseif(!empty($this->attributes[$field])){
						$attribs[$field] = htmlspecialchars($this->attributes[$field]);
					}
			}
		}
		return $attribs;
	}

}
