<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: f389fd111aab $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-02-17 01:17 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
class we_shop_shippingControl implements JsonSerializable{
	/** @var string */
	var $stateField = '';

	/** @var bool */
	var $isNet = true;

	/** @var int */
	var $vatId = 0;

	/** @var array<string,we_shop_shipping> */
	var $shippings = [];

	/** @var float */
	var $vatRate = 0;

	/**
	 * @param string $stateField
	 * @param bool $isNet
	 * @param int $vatId
	 * @param array<string,we_shop_shipping> $shippings
	 */
	private function __construct(string $stateField, bool $isNet, int $vatId, array $shippings){
		$this->stateField = $stateField;
		$this->isNet = $isNet;
		$this->vatId = $vatId;
		$this->shippings = $shippings;

		$this->vatRate = we_shop_vats::getVatRateForSite($vatId);
	}

	public static function getShippingControl(): self{
		$data = we_unserialize(fS('SELECT pref_value FROM ' . SETTINGS_TABLE . ' WHERE tool="shop" AND pref_name="weShippingControl"'));
		if($data){
			$shippingControl = self::initFromJson($data);
			$shippingControl->vatRate = we_shop_vats::getVatRateForSite($shippingControl->vatId);
			return $shippingControl;
		}
		return new self('', true, 1, []);
	}

	function setByRequest(): void{//FIXME: bad this is unchecked
		// this function inits a new entry, also it could change existing items
		$this->stateField = we_base_request::string('stateField');
		$this->isNet = we_base_request::bool('isNet');
		$this->vatId = we_base_request::int('vatId');

		$shipId = we_base_request::string('weShippingId');
		if(!$shipId){
			return;
		}

		$newShipping = new we_shop_shipping($shipId, we_base_request::string('weShipping_text'), we_base_request::stringList('weShipping_countries'), we_base_request::intList('weShipping_cartValue'), we_base_request::intList('weShipping_shipping'), we_base_request::bool('weShipping_default'));
		$this->shippings[$shipId] = $newShipping;

		if($newShipping->default){
			foreach($this->shippings as $id => &$shipping){
				if($id != $shipId){
					$shipping->default = false;
				}
			}
		}
	}

	function save(): bool{//FIXME: this must be changed to json!!
		return DB_WE::$inst->query('REPLACE INTO ' . SETTINGS_TABLE . ' SET ' . we_database_base::arraySetter([
					'tool' => 'shop',
					'pref_name' => 'weShippingControl',
					'pref_value' => we_serialize($this->jsonSerialize())
		]));
	}

	function delete(int $id): bool{
		if(isset($this->shippings[$id])){
			unset($this->shippings[$id]);
		}
		return $this->save();
	}

	function getShippingById(string $id): we_shop_shipping{
		return $this->shippings[$id];
	}

	function getDefaultShipping(): we_shop_shipping{
		foreach($this->shippings as $shipping){
			'@phan-var we_shop_shipping $shipping';
			if($shipping->default){
				return $shipping;
			}
		}

		throw new we_exception_shop('no default shipping found');
	}

	function getShippingCostByOrderValue(float $orderValue, array $customer = []): float{
		if($customer){
			// foreach, search the shipping
			if(isset($customer[$this->stateField])){
				foreach($this->shippings as $tmpShipping){
					if(in_array($customer[$this->stateField], $tmpShipping->countries, true)){
						$shipping = $tmpShipping;
						continue;
					}
				}
			}
		}
		try{
			$shipping = (empty($shipping) ? // take default shipping
				$this->getDefaultShipping() :
				$shipping);
		} catch (we_exception_shop $e){
			return 0;
		}

		$shippingId = 0;
		foreach($shipping->cartValue as $i => $cur){
			if($cur > $orderValue){
				continue;
			}
			$shippingId = $i;
		}
		return $shipping->shipping[$shippingId];
	}

	/**
	 *
	 * @return array<string,mixed>
	 */
	public function jsonSerialize()/* :mixed */{
		$ret = [
			'stateField' => $this->stateField,
			'isNet' => (bool) $this->isNet,
			'vatId' => (int) $this->vatId,
			'shippings' => [],
		];
		foreach($this->shippings as $id => $ship){
			$ret['shippings'][$id] = $ship->jsonSerialize();
		}
		return $ret;
	}

	public static function initFromJson(array $json): self{
		$ship = [];
		foreach($json['shippings'] as $id => $data){
			$ship[$id] = we_shop_shipping::initFromJson($data);
		}
		return new self($json['stateField'], (bool) $json['isNet'], (int) $json['vatId'], $ship);
	}

}
