<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 5a9d1f908018 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2020-02-08 13:08 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * General Definition of WebEdition Workflow
 *
 */
class we_workflow_workflow extends we_workflow_base{
	const INACTIVE = 0;
	const ACTIVE = 1;
// Document-Type/Category based Workflow Type
	const DOCTYPE_CATEGORY = 'doctype';
// Directory based Workflow Type
	const FOLDER = 'folder';
// Object based Workflow Type
	const OBJECT = 'object';

	//properties
	/** @var int */
	var $ID = 0;

	/** @var string */
	var $Text;

	/** @var string */
	var $Type = self::FOLDER;

	/** @var int[] */
	var $Folders = [0];

	/** @var int[] */
	var $DocType = [];

	/** @var int[] */
	var $Objects = [];

	/** @var int[] */
	var $Categories = [];

	/** @var int[] */
	var $ObjectFileFolders = [0];

	/** @var int[] */
	var $ObjCategories = [];

	/** @var int */
	var $active = self::INACTIVE;

	/** @var bool */
	var $EmailPath = false;

	/** @var bool */
	var $LastStepAutoPublish = false;

	/*	 * @var we_workflow_step[] steps for WorkFlow Definition 	 */
	var $steps = [];

	/** @var we_workflow_document default document object */
	var $documentDef;

	/** @var array documents array; format document[documentID]=document_name don't create array of objects 'cos want to save some memory */
	var $documents = [];

	/**
	 * Default Constructor
	 * Can load or create new Workflow Definition depends of parameter
	 */
	function __construct(int $workflowID = 0){
		parent::__construct(WORKFLOW_TABLE);

		$this->persistent_slots = [
			'ID' => we_base_request::INT,
			'Text' => we_base_request::STRING,
			'Type' => we_base_request::INT,
			'DocType' => we_base_request::INTLISTA,
			'Folders' => we_base_request::INTLISTA,
			'ObjectFileFolders' => we_base_request::INTLISTA,
			'Objects' => we_base_request::INTLISTA,
			'Categories' => we_base_request::INTLISTA,
			'ObjCategories' => we_base_request::INTLISTA,
			'active' => we_base_request::BOOL,
			'EmailPath' => we_base_request::BOOL,
			'LastStepAutoPublish' => we_base_request::BOOL,
		];

		$this->Text = g_l('modules_workflow', '[new_workflow]');

		$this->AddNewStep();
		$this->AddNewTask();

		if($workflowID){
			$this->load($workflowID);
		}
	}

	/**
	 * Load workflow definition from database
	 */
	function load(int $id = 0, bool $loadData = true): bool{
		parent::load($id);
		if(!$this->ID){
			return false;
		}

		// get steps for workflow
		$this->steps = we_workflow_step::getAllSteps($this->ID);
		$this->loadDocuments();
		return true;
	}

	/**
	 * get all documents for workflow from database
	 */
	function loadDocuments(): void{
		$docTable = ($this->Type === self::OBJECT ? OBJECT_FILES_TABLE : FILE_TABLE );
		$this->documents = $this->db->getAllq('SELECT w.ID,d.Text,d.ContentType AS contenttype
FROM ' .
			WORKFLOW_DOC_TABLE . ' w JOIN ' .
			$docTable . ' d ON w.documentID=d.ID
WHERE
	w.workflowID=' . intval($this->ID) . ' AND
Status="active"');
	}

	/**
	 * get all workflows from database (STATIC)
	 */
	function getAllWorkflows(): array{
		return $this->db->getAllq('SELECT ID FROM ' . WORKFLOW_TABLE . ' ORDER BY Text', true);
	}

	/**
	 * get all workflows from database
	 * @param int[] $all
	 * @return array<int,string>
	 */
	public static function getAllWorkflowsInfo(int $status = self::ACTIVE, array $types = [self::DOCTYPE_CATEGORY], array $all = []): array{
		//@phan-suppress-next-line PhanPartialTypeMismatchReturn
		return DB_WE::$inst->getAllFirstq('SELECT ID, Text FROM ' . WORKFLOW_TABLE . ' WHERE active = ' . $status . ' AND Type IN("' . implode('", "', $types) . '")' . ($all ? ' AND ID IN(' . implode(', ', $all) . ')' : '') . ' ORDER BY Text', false);
	}

	/**
	 * save complete workflow definition in database
	 * saving also all steps and tasks for current workflow
	 */
	function save(bool $force_new = false, bool $rebuild = false, array $skipWrite = []): bool{
		if(!parent::save()){
			return false;
		}

		// save all steps also

		$stepsList = [];

		foreach($this->steps as &$step){
			$step->workflowID = $this->ID;
			$step->save();

			$stepsList[] = $step->ID;
		}


		// !!! here we have to delete all other steps in database except this in array
		if($stepsList){
			$this->db->query('DELETE FROM ' . WORKFLOW_STEP_TABLE . ' WHERE workflowID = ' . intval($this->ID) . ' AND ID NOT IN (' . implode(', ', $stepsList) . ')');
		}

		//remove all documents from workflow
		foreach($this->documents as $val){
			$this->documentDef = new we_workflow_document($val['ID']);
			$this->documentDef->finishWorkflow(true);
			$this->documentDef->save();
		}
		return true;
	}

	/**
	 * delete workflow from database
	 * delete also all steps and tasks for current workflow
	 */
	function delete(): bool{
		if(!$this->ID){
			return false;
		}

		foreach(array_keys($this->steps) as $key){
			$this->steps[$key]->delete();
		}

		foreach($this->documents as $val){
			$this->documentDef = new we_workflow_document($val['ID']);
			$this->documentDef->delete();
		}

		parent::delete();

		return true;

		//$this->ID = -2; # status deleted
	}

	static function isDocInWorkflow(int $docID): bool{
		return fI('SELECT ID FROM ' . WORKFLOW_DOC_TABLE . ' WHERE documentID=' . $docID . ' AND Type IN("' . self::DOCTYPE_CATEGORY . '", "' . self::FOLDER . '") AND active=' . self::INACTIVE) > 0;
	}

	static function isObjectInWorkflow(int $docID): bool{
		return fI('SELECT ID FROM ' . WORKFLOW_DOC_TABLE . ' WHERE documentID=' . $docID . ' AND Type="' . self::OBJECT . '" AND active=' . self::INACTIVE) > 0;
	}

	/**
	 * Get workflow for document
	 * @param int $doctype
	 * @param int[] $categories
	 * @param int $folder
	 * @param int[] $all
	 * @return int
	 */
	static function getDocumentWorkflow(int $doctype, array $categories, int $folder, array &$all): int{
		/**
		 * find by document type (has to be together with category)
		 */
		$wfIDs = ($doctype ?
			DB_WE::$inst->getAllq('SELECT ID FROM ' . WORKFLOW_TABLE . ' WHERE FIND_IN_SET(' . $doctype . ', DocType) AND Type = ' . self::DOCTYPE_CATEGORY . ' AND active = ' . self::ACTIVE) :
			[]);

		/**
		 * find by category
		 */
		if($categories){
			foreach($categories as $v){
				if($doctype){
					$wfIDs = array_merge($wfIDs, DB_WE::$inst->getAllq('SELECT ID FROM ' . WORKFLOW_TABLE . ' WHERE DocType=' . $doctype . ' AND FIND_IN_SET(' . intval($v) . ', Categories) AND Type="' . self::DOCTYPE_CATEGORY . '" AND active=' . self::ACTIVE));
				} else {
					$wfIDs = array_merge($wfIDs, DB_WE::$inst->getAllq('SELECT ID FROM ' . WORKFLOW_TABLE . ' WHERE FIND_IN_SET(' . intval($v) . ', Categories) AND Type="' . self::DOCTYPE_CATEGORY . '" AND active=' . self::ACTIVE));
				}
			}
		}
		$wfIDs = array_count_values($wfIDs);
		$all = array_keys($wfIDs);
		'@phan-var int[] $all';
		return (self::findMaxWf($wfIDs) ?: // when we have found a document type-based workflow we can return
			(self::findWfIdForFolder($folder) ?: 0));
	}

	private static function findMaxWf(array $wfIDs): int{
		$workflowID = $max = 0;
		foreach($wfIDs as $wfID => $anz){
			if($anz > $max){
				$workflowID = $wfID;
				$max = $anz;
			}
		}
		return $workflowID;
	}

	private static function findWfIdForFolder(int $folderID): int{
		//FIXME: $all should be checked if present
		$folders = [$folderID];
		while($folderID != 0){
			$folderID = fI('SELECT ParentID FROM ' . FILE_TABLE . ' WHERE ID=' . $folderID);
			$folders[] = $folderID;
		}
		DB_WE::$inst->addTable('TMP_WF', ['ID' => 'INT unsigned NOT NULL'], [], 'MEMORY', true);
		DB_WE::$inst->query('INSERT INTO TMP_WF (ID) VALUES (' . implode('), (', $folders) . ')');
		$ret = fI('SELECT DISTINCT(w.ID) FROM ' . WORKFLOW_TABLE . ' w JOIN TMP_WF tmp ON FIND_IN_SET(tmp.ID, w.Folders) WHERE w.Type="' . self::FOLDER . '" AND w.active = ' . self::ACTIVE . ' LIMIT 1');
		DB_WE::$inst->delTable('TMP_WF', true);
		return $ret ?: 0;
	}

	/**
	 * Get workflow for object
	 * @param int $object
	 * @param int[] $categories
	 * @param int $folderID
	 * @param int[] $all
	 */
	public static function getObjectWorkflow(int $object, array $categories, int $folderID, array &$all): int{
		$wfIDs = DB_WE::$inst->getAllq('SELECT ID FROM ' . WORKFLOW_TABLE . ' WHERE FIND_IN_SET(' . $object . ', Objects) AND Type="' . self::OBJECT . '" AND active=' . self::ACTIVE . ($folderID ? ' AND (FIND_IN_SET(' . $folderID . ', ObjectFileFolders) OR FIND_IN_SET(0, ObjectFileFolders) OR ObjectFileFolders = "")' : ''));

		/**
		 * find by category
		 */
		if($categories){
			foreach($categories as $v){
				$wfIDs = array_merge($wfIDs, DB_WE::$inst->getAllq('SELECT ID FROM ' . WORKFLOW_TABLE . ' WHERE FIND_IN_SET(' . $object . ', Objects) AND FIND_IN_SET(' . intval($v) . ', ObjCategories) AND Type="' . self::OBJECT . '" AND active=' . self::ACTIVE));
			}
		}
		$wfIDs = array_count_values($wfIDs);
		$all = array_keys($wfIDs);
		'@phan-var int[] $all';

		return (self::findMaxWf($wfIDs) ?: 0);
	}

	function addNewStep(): void{
		$this->steps[] = new we_workflow_step();
	}

	function addNewTask(): void{
		foreach($this->steps as &$v){
			$v->tasks[] = new we_workflow_task();
		}
	}

	public static function getJSLangConsts(): string{
		return we_base_util::setLangString([
				'view' => [
					'delete_question' => g_l('modules_workflow', '[delete_question]'),
					'emty_log_question' => g_l('modules_workflow', '[emty_log_question]'),
					'nothing_to_delete' => (g_l('modules_workflow', '[nothing_to_delete]')),
					'nothing_to_save' => (g_l('modules_workflow', '[nothing_to_save]')),
					'save_changed_workflow' => g_l('modules_workflow', '[save_changed_workflow]'),
					'save_question' => g_l('modules_workflow', '[save_question]'),
				],
				'prop' => [
					'del_last_step' => (g_l('modules_workflow', '[del_last_step]')),
					'del_last_task' => (g_l('modules_workflow', '[del_last_task]')),
					'doctype_empty' => (g_l('modules_workflow', '[doctype_empty]')),
					'folders_empty' => (g_l('modules_workflow', '[folders_empty]')),
					'name_empty' => (g_l('modules_workflow', '[name_empty]')),
					'objects_empty' => (g_l('modules_workflow', '[objects_empty]')),
					'user_empty' => (g_l('modules_workflow', '[user_empty]')),
					'worktime_empty' => (g_l('modules_workflow', '[worktime_empty]')),
				]
		]);
	}

}
